package net.qiujuer.genius.graphics;

import android.graphics.Bitmap;

/**
 * In this can blurring your image by {@link Bitmap}
 * The Blurring support:
 * {@link android.graphics.Bitmap.Config#RGB_565},
 * {@link android.graphics.Bitmap.Config#ARGB_8888}
 */
@SuppressWarnings("JavaDoc")
public final class Blur {

    /**
     * Building the bitmap
     *
     * @param original Bitmap
     * @return Bitmap
     */
    private static Bitmap checkSource(Bitmap original, int radius) {
        if (radius < 0 || radius > 256) {
            throw new RuntimeException("Blur bitmap radius must >= 1 and <=256.");
        }

        // First we should check the original
        if (original == null) {
            throw new NullPointerException("Blur bitmap original isn't null.");
        }
        if (original.isRecycled()) {
            throw new RuntimeException("Blur bitmap can't blur a recycled bitmap.");
        }

        Bitmap.Config config = original.getConfig();
        if (config != Bitmap.Config.ARGB_8888 && config != Bitmap.Config.RGB_565) {
            throw new RuntimeException("Blur bitmap only supported Bitmap.Config.ARGB_8888 and Bitmap.Config.RGB_565.");
        }

        return (original);
    }

    /**
     * StackBlur By Jni Bitmap
     *
     * @param original Original Image
     * @param radius   Blur radius
     * @return Image Bitmap
     */
    public static Bitmap onStackBlur(Bitmap original, int radius) {
        Bitmap bitmap = checkSource(original, radius);

        // Return this none blur
        if (radius == 1) {
            return bitmap;
        }

        //Jni BitMap Blur
        nativeStackBlurBitmap(bitmap, radius);

        return (bitmap);
    }


    /**
     * StackBlur By Jni Bitmap
     * in this we will cut the source bitmap to some parts.
     * We'll deal(blur) with it one by one. This will reduce the memory consumption.
     *
     * @param original Original Image
     * @param radius   Blur radius
     * @return Image Bitmap
     */
    public static Bitmap onStackBlurClip(Bitmap original, int radius) {
        Bitmap bitmap = checkSource(original, radius);

        // Return this none blur
        if (radius == 1) {
            return bitmap;
        }

        int h = bitmap.getHeight();
        int w = bitmap.getWidth();

        final int clipPixels = 1024 * 256;
        float clipScale = (h * w) / clipPixels;
        int minLen = radius + radius + 50;

        if (clipScale >= 2) {
            float itemLen = h / clipScale;
            itemLen = itemLen < minLen ? minLen : itemLen;
            clipScale = h / itemLen;
        }

        if (clipScale < 2) {
            //Jni BitMap Blur
            nativeStackBlurBitmap(bitmap, radius);
        } else {
            if (clipScale > 12)
                clipScale = 12;
            //Jni BitMap Blur
            onStackBlurClip(bitmap, radius, (int) clipScale);
        }

        return (bitmap);
    }

    /**
     * StackBlur By Jni Bitmap
     * in this we will cut the source bitmap to some parts.
     * We'll deal(blur) with it one by one. This will reduce the memory consumption.
     *
     * @param original Original Image
     * @param radius   Blur radius
     * @param parts    Blur cut parts count
     * @return Image Bitmap
     */
    public static Bitmap onStackBlurClip(Bitmap original, int radius, int parts) {
        Bitmap bitmap = checkSource(original, radius);
        if (parts < 2 || parts > 12) {
            throw new RuntimeException("Blur bitmap parts must >= 2 and <=12.");
        }

        if (original.getHeight() / parts < radius + radius) {
            throw new RuntimeException("Blur bitmap height/partsCount must > radius+radius values.");
        }

        //Jni BitMap Blur
        nativeStackBlurBitmapClip(bitmap, radius, parts);

        return (bitmap);
    }

    /**
     * StackBlur By Jni Pixels
     *
     * @param pix    Original Image, you can call:
     *               <p>
     *               int w = bitmap.getWidth();
     *               int h = bitmap.getHeight();
     *               int[] pix = new int[w * h];
     *               bitmap.getPixels(pix, 0, w, 0, 0, w, h);
     *               <p>
     *               // Jni Pixels Blur
     *               onStackBlurPixels(pix, w, h, radius);
     *               <p>
     *               bitmap.setPixels(pix, 0, w, 0, 0, w, h);
     *               <p>
     * @param radius Blur radius
     * @return Image Bitmap
     */
    public static int[] onStackBlurPixels(int[] pix, int w, int h, int radius) {
        if (radius < 0 || radius > 256) {
            throw new RuntimeException("Blur bitmap radius must >= 1 and <=256.");
        }

        if (pix == null) {
            throw new RuntimeException("Blur bitmap pix isn't null.");
        }

        if (pix.length < w * h) {
            throw new RuntimeException("Blur bitmap pix length must >= w * h.");
        }

        // Jni Pixels Blur
        nativeStackBlurPixels(pix, w, h, radius);

        return (pix);
    }

    /**
     * StackBlur By Java Bitmap
     *
     * @param original Original Image
     * @param radius   Blur radius
     * @return Image Bitmap
     */
    public static Bitmap onStackBlurJava(Bitmap original, int radius) {
        // Stack Blur v1.0 from
        // http://www.quasimondo.com/StackBlurForCanvas/StackBlurDemo.html
        //
        // Java Author: Mario Klingemann <mario at quasimondo.com>
        // http://incubator.quasimondo.com
        // created Feburary 29, 2004
        // Android port : Yahel Bouaziz <yahel at kayenko.com>
        // http://www.kayenko.com
        // ported april 5th, 2012

        // This is a compromise between Gaussian Blur and Box blur
        // It creates much better looking blurs than Box Blur, but is
        // 7x faster than my Gaussian Blur implementation.
        //
        // I called it Stack Blur because this describes best how this
        // filter works internally: it creates a kind of moving stack
        // of colors whilst scanning through the image. Thereby it
        // just has to add one new block of color to the right side
        // of the stack and remove the leftmost color. The remaining
        // colors on the topmost layer of the stack are either added on
        // or reduced by one, depending on if they are on the right or
        // on the left side of the stack.
        //
        // If you are using this algorithm in your code please add
        // the following line:
        //
        // Stack Blur Algorithm by Mario Klingemann <mario@quasimondo.com>

        Bitmap bitmap = checkSource(original, radius);

        // Return this none blur
        if (radius == 1) {
            return bitmap;
        }

        int w = bitmap.getWidth();
        int h = bitmap.getHeight();

        int[] pix = new int[w * h];
        // get array
        bitmap.getPixels(pix, 0, w, 0, 0, w, h);

        // run Blur
        int wm = w - 1;
        int hm = h - 1;
        int wh = w * h;
        int div = radius + radius + 1;

        short r[] = new short[wh];
        short g[] = new short[wh];
        short b[] = new short[wh];
        int rSum, gSum, bSum, x, y, i, p, yp, yi, yw;
        int vMin[] = new int[Math.max(w, h)];

        int divSum = (div + 1) >> 1;
        divSum *= divSum;

        short dv[] = new short[256 * divSum];
        for (i = 0; i < 256 * divSum; i++) {
            dv[i] = (short) (i / divSum);
        }

        yw = yi = 0;

        int[][] stack = new int[div][3];
        int stackPointer;
        int stackStart;
        int[] sir;
        int rbs;
        int r1 = radius + 1;
        int routSum, goutSum, boutSum;
        int rinSum, ginSum, binSum;

        for (y = 0; y < h; y++) {
            rinSum = ginSum = binSum = routSum = goutSum = boutSum = rSum = gSum = bSum = 0;
            for (i = -radius; i <= radius; i++) {
                p = pix[yi + Math.min(wm, Math.max(i, 0))];
                sir = stack[i + radius];
                sir[0] = (p & 0xff0000) >> 16;
                sir[1] = (p & 0x00ff00) >> 8;
                sir[2] = (p & 0x0000ff);

                rbs = r1 - Math.abs(i);
                rSum += sir[0] * rbs;
                gSum += sir[1] * rbs;
                bSum += sir[2] * rbs;
                if (i > 0) {
                    rinSum += sir[0];
                    ginSum += sir[1];
                    binSum += sir[2];
                } else {
                    routSum += sir[0];
                    goutSum += sir[1];
                    boutSum += sir[2];
                }
            }
            stackPointer = radius;

            for (x = 0; x < w; x++) {

                r[yi] = dv[rSum];
                g[yi] = dv[gSum];
                b[yi] = dv[bSum];

                rSum -= routSum;
                gSum -= goutSum;
                bSum -= boutSum;

                stackStart = stackPointer - radius + div;
                sir = stack[stackStart % div];

                routSum -= sir[0];
                goutSum -= sir[1];
                boutSum -= sir[2];

                if (y == 0) {
                    vMin[x] = Math.min(x + radius + 1, wm);
                }
                p = pix[yw + vMin[x]];

                sir[0] = (p & 0xff0000) >> 16;
                sir[1] = (p & 0x00ff00) >> 8;
                sir[2] = (p & 0x0000ff);

                rinSum += sir[0];
                ginSum += sir[1];
                binSum += sir[2];

                rSum += rinSum;
                gSum += ginSum;
                bSum += binSum;

                stackPointer = (stackPointer + 1) % div;
                sir = stack[(stackPointer) % div];

                routSum += sir[0];
                goutSum += sir[1];
                boutSum += sir[2];

                rinSum -= sir[0];
                ginSum -= sir[1];
                binSum -= sir[2];

                yi++;
            }
            yw += w;
        }
        for (x = 0; x < w; x++) {
            rinSum = ginSum = binSum = routSum = goutSum = boutSum = rSum = gSum = bSum = 0;
            yp = -radius * w;
            for (i = -radius; i <= radius; i++) {
                yi = Math.max(0, yp) + x;

                sir = stack[i + radius];

                sir[0] = r[yi];
                sir[1] = g[yi];
                sir[2] = b[yi];

                rbs = r1 - Math.abs(i);

                rSum += r[yi] * rbs;
                gSum += g[yi] * rbs;
                bSum += b[yi] * rbs;

                if (i > 0) {
                    rinSum += sir[0];
                    ginSum += sir[1];
                    binSum += sir[2];
                } else {
                    routSum += sir[0];
                    goutSum += sir[1];
                    boutSum += sir[2];
                }

                if (i < hm) {
                    yp += w;
                }
            }
            yi = x;
            stackPointer = radius;
            for (y = 0; y < h; y++) {
                // Preserve alpha channel: ( 0xff000000 & pix[yi] )
                pix[yi] = (0xff000000 & pix[yi]) | (dv[rSum] << 16) | (dv[gSum] << 8) | dv[bSum];

                rSum -= routSum;
                gSum -= goutSum;
                bSum -= boutSum;

                stackStart = stackPointer - radius + div;
                sir = stack[stackStart % div];

                routSum -= sir[0];
                goutSum -= sir[1];
                boutSum -= sir[2];

                if (x == 0) {
                    vMin[y] = Math.min(y + r1, hm) * w;
                }
                p = x + vMin[y];

                sir[0] = r[p];
                sir[1] = g[p];
                sir[2] = b[p];

                rinSum += sir[0];
                ginSum += sir[1];
                binSum += sir[2];

                rSum += rinSum;
                gSum += ginSum;
                bSum += binSum;

                stackPointer = (stackPointer + 1) % div;
                sir = stack[stackPointer];

                routSum += sir[0];
                goutSum += sir[1];
                boutSum += sir[2];

                rinSum -= sir[0];
                ginSum -= sir[1];
                binSum -= sir[2];

                yi += w;
            }
        }

        // set Bitmap
        bitmap.setPixels(pix, 0, w, 0, 0, w, h);

        return (bitmap);
    }


    static {
        System.loadLibrary("genius_graphics");
    }

    /**
     * Blur Image By Pixels
     *
     * @param pixels Img pixel array
     * @param w      Img width
     * @param h      Img height
     * @param r      Blur radius
     * @hide
     */
    private static native void nativeStackBlurPixels(int[] pixels, int w, int h, int r);

    /**
     * Blur Image By Bitmap
     *
     * @param bitmap Img Bitmap
     * @param r      Blur radius
     * @hide
     */
    private static native void nativeStackBlurBitmap(Bitmap bitmap, int r);

    /**
     * Blur image by bitmap,
     * in this we will cut the source bitmap to some parts.
     * We'll deal(blur) with it one by one. This will reduce the memory consumption.
     *
     * @param bitmap Img Bitmap
     * @param r      Blur radius
     * @param parts  Cut the image to parts by the vertical orientation
     * @hide
     */
    private static native void nativeStackBlurBitmapClip(Bitmap bitmap, int r, int parts);
}
